#!/usr/bin/perl
#    Original VB script: 
#    Copyright (C) 2009 Ian Richardson, Murray Thomson
#    CREST (Centre for Renewable Energy Systems Technology),
#    Department of Electronic and Electrical Engineering
#    Loughborough University, Leicestershire LE11 3TU, UK
#    Tel. +44 1509 635326. Email address: I.W.Richardson2@lboro.ac.uk

#    Perl implementation:
#    Copyright (C) 2010 Gergely Acs <acs@crysys.hu>

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
package ApplianceModel;
use strict;
use Switch;
use Date::Calc qw(Add_Delta_Days); 
use ProbabilityModifier;
use vars qw($VERSION);
$VERSION = '0.01';

# Create a collection to store the activity statisitics
my %oActivityStatistics;

# Declare the appliance description variables
my $sApplianceType;
my $iMeanCycleLength;
my $iCyclesPerYear;
my $iStandbyPower;
my $iRatedPower;
my $dCalibration;
my $dOwnership;
my $iTargetAveragekWhYear;
my $sUseProfile;
my $iRestartDelay;
my $bHasAppliance;

# Declare timing counter and current power use variables
my $iCycleTimeLeft;
my $iRestartDelayTimeLeft;
my $iPower;

sub new 
{
	my $package = shift;
	%oActivityStatistics = ();
	return bless({}, $package);
}

sub RunApplianceSimulation
{
    my $self = shift;
    # Get the month
    my ($tmp, $iMonth, $bWeekend, $Lighting_Model, $occ_sim_data, $appliance_sim_data, $activity_stats) = @_;

    my @appliances = @{$tmp};

    # Declare variables
    my $dRan;
    my $dActivityProbability;
    my $iTenMinuteCount;
    my $iAppliance;
    my $iMinute;
    my $iActiveOccupants;
    my $sKey;
    # my $sCell;
    #my $iApplianceSourceCellOffsetY;
    #my $iApplianceTargetCellOffsetX;
    
    # Declare an array to store an annual temperature profile (used for heating appliances)
    my $oMonthlyRelativeTemperatureModifier;
    
    # Define the relative monthly temperatures
    # Data derived from MetOffice temperature data for the Midlands in 2007 (http://www.metoffice.gov.uk/climate/uk/2007/) Crown Copyright
    my @oMonthlyRelativeTemperatureModifier = (0, 1.63, 1.821, 1.595, 0.867, 0.763, 0.191, 0.156, 0.087, 
		    0.399, 0.936, 1.561, 1.994);
    
    # Define the cell count offsets
    #my $iApplianceSourceCellOffsetY = 12;
    #my $iApplianceTargetCellOffsetX = 7;
    
    # Load the activity statistics into a collection
    $self->LoadActivityStatistics($activity_stats);
    
    # Declare an array to store the simulation data
    my @vSimulationArray; #(1 To 1442, 1 To 33) As Variant
    
    # For each of the appliances
    foreach my $appliance (@appliances) 
    #(my $iAppliance = 0; $iAppliance < 33; $iAppliance++)
    {
        # Initialisation
        $iCycleTimeLeft = 0;
        $iRestartDelayTimeLeft = 0;
    
        # Get the appliance details
	$iAppliance = $appliance->ID;
        $sApplianceType = $appliance->Type; 
	#Range("'appliances'!T" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value
        $iMeanCycleLength = $appliance->MeanCycleLength; 
	# Range("'appliances'!H" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value
        $iCyclesPerYear = $appliance->CyclesPerYear;  
	# Range("'appliances'!G" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value
        $iStandbyPower = $appliance->StandbyPower; 
	# Range("'appliances'!J" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value
        $iRatedPower = $appliance->RatedPower; 
	# Range("'appliances'!I" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value
        $dCalibration = $appliance->Calibration; 
	# Range("'appliances'!W" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value
        $dOwnership = $appliance->Ownership; 
	# Range("'appliances'!E" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value
        $iTargetAveragekWhYear = $appliance->TargetAveragekWhYear; 
	# Range("'appliances'!Z" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value
        $sUseProfile = $appliance->UserProfile; 
	# Range("'appliances'!U" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value
        $iRestartDelay = $appliance->RestartDelay; 
	# Range("'appliances'!K" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value
        $bHasAppliance = $appliance->HasAppliance; 
	# IIf(Range("'appliances'!D" + CStr(iAppliance + iApplianceSourceCellOffsetY)).Value = "YES", True, False)
        
        # Write the appliance type into the target sheet header
        $vSimulationArray[0][$iAppliance] = $appliance->Name;
        
        # Write the units
        #$vSimulationArray[1][$iAppliance] = "(W)";
            
        # Check if this appliance is assigned to this dwelling
        if (!$bHasAppliance)
	{
            # This appliance is not applicable, so write zeros to the power demand
            for (my $iCount = 1; $iCount < 1441; $iCount++)
	    {
                $vSimulationArray[$iCount][$iAppliance] = 0;
	    }
	    next;
        }
         
        # Randomly delay the start of appliances that have a restart delay (e.g. cold appliances with more regular intervals)
        $iRestartDelayTimeLeft = rand() * $iRestartDelay * 2; # Weighting is 2 just to provide some diversity
        
	# Make the rated power variable over a normal distribution to provide some variation
        $iRatedPower = LightingModel::GetMonteCarloNormalDistGuess($iRatedPower, $iRatedPower / 10); 
    
        # Get the weekday or weekend flag
        #$bWeekend = IIf(Range("K5").Value = "wd", False, True); 
             
        # Loop through each minute of the day
        $iMinute = 0;
        while ($iMinute < 1440)
	{
                # Set the default (standby) power demand at this time step
                $iPower = $iStandbyPower;
            
                # Get the ten minute period count
                $iTenMinuteCount = int($iMinute / 10);
            
                # Get the number of current active occupants for this minute
                # Convert from 10 minute to 1 minute resolution
                $iActiveOccupants = $occ_sim_data->GetActiveOccupants($iTenMinuteCount); 
		#Range("occ_sim_data!C" + CStr(11 + $iTenMinuteCount)).Value;
    
                # Now generate a key to get the activity statistics
                $sKey = $bWeekend . "_" . $iActiveOccupants . "_" . $sUseProfile; 
		#IIf(bWeekend, "1", "0") + "_" + CStr(iActiveOccupants) + "_" + sUseProfile ;
    
                # If this appliance is off having completed a cycle (ie. a restart delay)
                if ($iCycleTimeLeft <= 0 && $iRestartDelayTimeLeft > 0) 
		{
                    # Decrement the cycle time left
                    $iRestartDelayTimeLeft--;
		}
                # Else if this appliance is off
                elsif ($iCycleTimeLeft <= 0) 
		{
                    # There must be active occupants, or the profile must not depend on occupancy for a start event to occur
                    if ( ($iActiveOccupants > 0 && $sUseProfile ne "CUSTOM") || 
				    $sUseProfile eq "LEVEL") 
		    {
    
                        # Variable to store the event probability (default to 1)
                        $dActivityProbability = 1;
                        
                        # For appliances that depend on activity profiles and is not a custom profile ...
                        if ($sUseProfile ne "LEVEL" && $sUseProfile ne "ACTIVE_OCC" 
					&& $sUseProfile ne "CUSTOM") 
			{
                            # Get the probability for this activity profile for this time step
                            $dActivityProbability = $oActivityStatistics{$sKey}->GetModifiers($iTenMinuteCount);
                        
                        # For electric space heaters ... (excluding night storage heaters)
			}
                        elsif ($sApplianceType eq "ELEC_SPACE_HEATING") 
			{
                            # If this appliance is an electric space heater, then then activity probability is a function of the month of the year
                            $dActivityProbability = $oMonthlyRelativeTemperatureModifier[$iMonth-1];
                        
                        }
    
                        # Check the probability of a start event

			#	if ($sApplianceType eq "MICROWAVE" || $sApplianceType eq "OVEN")
			#	{
			#	print "Probbbbbb: $dActivityProbability\n";
			#	print "calib: $dCalibration\n";
			#	print "threshold: ", $dCalibration * $dActivityProbability,"\n";
			#	}
                        if (rand() < $dCalibration * $dActivityProbability) 
			{
#				if ($sApplianceType eq "MICROWAVE")
#				{
#				print "Suotes be!!!!!!\n";
#				}
                            # This is a start event
                            $self->StartAppliance();
			}
		    }
                    # Custom appliance handler: storage heaters have a simple representation
                    elsif ($sUseProfile eq "CUSTOM" && $sApplianceType eq "STORAGE_HEATER") 
		    {
                        # The number of cycles (one per day) set out in the calibration sheet
                        # is used to determine whether the storage heater is used
                        
                        # This model does not account for the changes in the Economy 7 time
                        # It assumes that the time starts at 00:30 each day
                        if ($iTenMinuteCount == 3) #  ie. 00:30 - 00:40
			{
                        
                            # Assume January 14th is the coldest day of the year
                            # oDate = #1/14/1997#
                            
                            # Get the month and day when the storage heaters are turned on and off, using the number of cycles per year
                          #  $oDateOff = DateAdd("d", $iCyclesPerYear / 2, $oDate);
			    my ($oYearOff, $iMonthOff, $iDayOff) = Add_Delta_Days(1997, 1, 14, int($iCyclesPerYear / 2)); 
			    my ($oYearOn, $iMonthOn, $iDayOn) = Add_Delta_Days(1997, 1, 14, -int($iCyclesPerYear / 2)); 
			    #print ("Storage heater, date on: $oYearOn/$iMonthOn/$iDayOn\n");
			    #print ("Storage heater, date off: $oYearOff/$iMonthOff/$iDayOff\n");

                            #$oDateOn = DateAdd("d", -$iCyclesPerYear / 2, $oDate);
                            #$iMonthOff = DatePart("m", $oDateOff);
                            #$iMonthOn = DatePart("m", $oDateOn);
                            
                            # Declare a probability of use variable
                            my $dProbability;
                            
                            # If this is a month in which the appliance is turned on of off
                            if ($iMonth == $iMonthOff || $iMonth == $iMonthOn) 
			    {
                                # Pick a 50% chance since this month has only a month of year resolution
                                $dProbability = 0.5 / 10;  # (since there are 10 minutes in this period)
			    }
                            elsif ($iMonth > $iMonthOff && $iMonth < $iMonthOn)
			    {
                                # The appliance is not used in summer
                                $dProbability = 0
			    }
                            else
			    {
                                # The appliance is used in winter
                                $dProbability = 1
                            }
                                                      
                            # Determine if a start event occurs
                            if (rand() <= $dProbability)
			    {
                                # This is a start event
                                $self->StartAppliance();
			    }
                                
                         }
		    }
                }
                else
		{
                    # The appliance is on - if the occupants become inactive, switch off the appliance
                    if ($iActiveOccupants == 0 && $sUseProfile ne "LEVEL" && 
				    $sUseProfile ne "ACT_LAUNDRY" && $sUseProfile ne "CUSTOM")
		    {
                             
                        # Do nothing. The activity will be completed upon the return of the active occupancy.
                        # Note that LEVEL means that the appliance use is not related to active occupancy.
                        # Note also that laundry appliances do not switch off upon a transition to inactive occupancy.
		    }
                    else
		    {
                        # Set the power
                        $iPower = $self->GetPowerUsage($iCycleTimeLeft);
                        
                        # Decrement the cycle time left
                        $iCycleTimeLeft--;
			$appliance_sim_data->AddActiveAppliance($iMinute, $appliance->Type, $appliance->Name);
                    }
		}
    
                # Set the appliance power at this time step
                $vSimulationArray[1 + $iMinute][$iAppliance] = $iPower;
            
                # Increment the time
                $iMinute++;
                
          }  
       }        
    
    # Write the data back to the simulation sheet
    #Range("appliance_sim_data!H10:AN1451") = vSimulationArray;
    $appliance_sim_data->SimulationArray(\@vSimulationArray);
}
    
sub CycleLength
{
    # Set the value to that provided in the configuration
    my $ret = $iMeanCycleLength;
    
    # Use the TV watching length data approximation, derived from the TUS data
    if ($sApplianceType eq "TV1" || $sApplianceType eq "TV2" || $sApplianceType eq "TV3")
    {
        # The cycle length is approximated by the following function
        # The avergage viewing time is approximately 73 minutes
        $ret = int(70 * ((0 - log(1 - rand())) ** 1.1));
    }
    # BUG ??????????
    elsif ($sApplianceType eq "STORAGE_HEATER" || $sApplianceType eq "ELEC_SPACE_HEATING")
    {
        # Provide some variation on the cycle length of heating appliances
        $ret = LightingModel::GetMonteCarloNormalDistGuess($iMeanCycleLength, $iMeanCycleLength / 10);
    }    
    return $ret;
}

sub GetPowerUsage #(ByVal iCycleTimeLeft As Integer)
{
    # Set the return power to the rated power
    shift;
    my $iCycleTimeLeft = $_[0];
    my $ret = $iRatedPower;

    # Some appliances have a custom (variable) power profile depending on the time left
    if ($sApplianceType eq "WASHING_MACHINE" || $sApplianceType eq "WASHER_DRYER")
    {
        
            # Declare the total cycle time variable
            my $iTotalCycleTime;
        
            # Calculate the washing cycle time
            $iTotalCycleTime = 138 if $sApplianceType eq "WASHING_MACHINE";
            $iTotalCycleTime = 198 if $sApplianceType eq "WASHER_DRYER";
        
            # This is an example power profile for an example washing machine
            # This simplistic model is based upon data from personal communication with a major washing maching manufacturer
            switch ($iTotalCycleTime - $iCycleTimeLeft + 1)
	    {
                case [1 .. 8] { $ret = 73 }         # Start-up and fill
                case [9 .. 29] { $ret = 2056 }      # Heating
                case [30 .. 81] { $ret = 73 }       # Wash and drain
                case [82 .. 92] { $ret = 73 }       # Spin
                case [93 .. 94] { $ret = 250 }      # Rinse
                case [95 .. 105] { $ret = 73 }     # Spin
                case [106 .. 107] { $ret = 250 }    # Rinse
                case [108 .. 118] { $ret = 73 }     # Spin
                case [119 .. 120] { $ret = 250 }    # Rinse
                case [121 .. 131] { $ret = 73 }     # Spin
                case [132 .. 133] { $ret = 250 }    # Rinse
                case [134 .. 138] { $ret = 568 }   # Fast spin
                case [139 .. 198] { $ret = 2500 }   # Drying cycle
                else  { $ret = $iStandbyPower }
	    }
    } 
    return $ret;
}


# Load the activity statistics into a collection
sub LoadActivityStatistics
{
    # Declare the variables
    shift;
    my $sKey;
    #my $sCell;
    my $activity_stats = $_[0];
    
    # Load the activity statistics into probability modifiers objects
    my $activity_num = $activity_stats->GetNumberOfActivites();
    for (my $i = 0; $i < $activity_num; $i++)
    {
        # Create a new probability modifier
        my $oActivityStatsItem = new ProbabilityModifier();
    
        # Read in the data
        $oActivityStatsItem->IsWeekend($activity_stats->GetWeekendFlag($i)); 
	#IIf(Range("'activity_stats'!B" + CStr(i)).Value = 1, True, False)
        $oActivityStatsItem->ActiveOccupantCount($activity_stats->GetActiveOccupantCount($i)); 
	#Range("'activity_stats'!C" + CStr(i)).Value
        $oActivityStatsItem->ID($activity_stats->GetName($i)); 
	#$Range("'activity_stats'!D" + CStr(i)).Value
        
        # Get the hourly modifiers
        for (my $j = 0; $j < 144; $j++)
	{
            # Get the column reference
            #Cells(i, j + 5).Address(True, False, xlA1)
        
            # Read the values
            $oActivityStatsItem->SetModifiers($j, $activity_stats->GetProportion($i, $j)); 
	    #Range("'activity_stats'!" + sCell).Value
	}

        # Now generate a key
        #$sKey = IIf(oActivityStatsItem.IsWeekend, "1", "0") + "_" + CStr(oActivityStatsItem.ActiveOccupantCount) + "_" + oActivityStatsItem.ID
        $sKey = $oActivityStatsItem->IsWeekend . "_" . $oActivityStatsItem->ActiveOccupantCount 
	. "_" . uc($oActivityStatsItem->ID);
        
        # Add this object to the collection
	$oActivityStatistics{$sKey} = $oActivityStatsItem; 
#		oActivityStatistics.Add Item:=oActivityStatsItem, Key:=sKey
    }    
}


sub ConfigureAppliancesInDwelling
{
    shift;
    my @appliances = @{$_[0]};
    
    # For each appliance
    foreach my $app (@appliances)
    {
        # Determine if this simulated house has this appliance
	$app->HasAppliance(rand() < $app->Ownership);
    }    
}

# Start a cycle for the current appliance
sub StartAppliance
{
    my $self = shift;
    # Determine how long this appliance is going to be on for
    $iCycleTimeLeft = $self->CycleLength();
    
    # Determine if this appliance has a delay after the cycle before it can restart
    $iRestartDelayTimeLeft = $iRestartDelay;
    
    # Set the power
    $iPower = $self->GetPowerUsage($iCycleTimeLeft);
    
    # Decrement the cycle time left
    $iCycleTimeLeft--;
}
                            
1;
