#!/usr/bin/perl
#    Original VB script: 
#    Copyright (C) 2009 Ian Richardson, Murray Thomson
#    CREST (Centre for Renewable Energy Systems Technology),
#    Department of Electronic and Electrical Engineering
#    Loughborough University, Leicestershire LE11 3TU, UK
#    Tel. +44 1509 635326. Email address: I.W.Richardson2@lboro.ac.uk

#    Perl implementation:
#    Copyright (C) 2010 Gergely Acs <acs@crysys.hu>

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

package ApplianceSimData;
use strict;
use List::Util 'sum';
use List::MoreUtils qw(pairwise);
use Chart::Graph::Gnuplot qw(&gnuplot);
use accessors qw( SimulationArray );
use vars qw($VERSION);
$VERSION = '0.01';

my @active_appliances;
my @lighting_consumption;
my %used_appliances;

sub new 
{
	my $package = shift;

	@active_appliances = ();
	@lighting_consumption = ();
	%used_appliances = ();

	return bless({}, $package);
}

sub GetSumConsumption
{
	my $self = shift;
	my @arr = @{$self->SimulationArray};
	my @ret;

	for (my $i = 1; $i < scalar(@arr); $i++)
	{
		push @ret, List::Util::sum(@{$arr[$i]});
	}

	return @ret;
}

sub GetUsedAppliances
{
	return keys %used_appliances;
}

sub GetSlots
{
	shift;
	my $app = $_[0];

	my $last = -1;
	my $start;
	my @slots;
#	print "$app: @{$used_appliances{$app}}\n";
	foreach my $time (sort {$a <=> $b} @{$used_appliances{$app}})
	{
		if ($last == -1)
		{
			$start = $time 
		}
		elsif ($time - $last != 1 && $time != $last)
		{
			push(@slots, [$start, $last]);
			$start = $time;
		}

		$last = $time;
	}
	push(@slots, [$start, $last]) if $last != -1;
	return @slots;
}

sub AddActiveAppliance
{
	shift;

	my ($time, $app, $app_long) = @_;

	push @{$active_appliances[$time]}, $app;
	push @{$used_appliances{$app_long}}, $time;
}

sub WriteToFile
{
	my $self = shift;

	my ($filename1, $filename2, $filename3, $ref) = @_;
	my @lighting_consumption = @{$ref};

	die("Unable to open $filename1!") unless open (OUT, ">" . $filename1);

	my @arr = @{$self->SimulationArray};

	$" = ",";
	for (my $i = 0; $i < scalar(@arr); $i++)
	{
		print OUT ($i > 0 ? $i : ''), ",", ($i > 0 ? List::Util::sum(@{$arr[$i]}) : 'Sum') . 
			",@{$arr[$i]}\n"; 
	}

	close OUT;

	die("Unable to open $filename2!") unless open (OUT, ">" . $filename2);

	$" = ",";
	for (my $i = 0; $i < scalar(@active_appliances); $i++)
	{
		print OUT $i+1, ",", (defined $active_appliances[$i] ?  "@{$active_appliances[$i]}" : "None"), "\n"; 
	}

	close OUT;

	die("Unable to open $filename3!") unless open (OUT, ">" . $filename3);

	my @appliance_consumption = $self->GetSumConsumption();
	my @all_consumption = pairwise { $a + $b } @appliance_consumption, @lighting_consumption;

	$" = ",";
	for (my $i = 0; $i < scalar(@all_consumption); $i++)
	{
		print OUT ($i + 1) . ",$all_consumption[$i]\n"; 
	}

	close OUT;

}

sub AddLighting
{
	my $self = shift;

	@lighting_consumption = @_;
# TODO
	for (my $i = 0; $i < scalar(@lighting_consumption); $i++)
	{
		$self->AddActiveAppliance($i, "LIGHTING", "Lighting") 
			if $lighting_consumption[$i] > 0;
	}
}

sub MakePlots
{
	my $self = shift;
	my ($consumption_filename, $consumption_title, $usage_filename, $usage_title) = @_;

# Getting summary for plots
	my @appliance_consumption = $self->GetSumConsumption();

	my @all_consumption = pairwise { $a + $b } @appliance_consumption, @lighting_consumption;
	my @seconds = map {$_ * 60} (0..1439);

	gnuplot({'title' => $consumption_title,
				'output type' => 'eps',
				'output file' => $consumption_filename, 
				'xdata' => 'time',
				'timefmt' => '%s',
				'format' => ['x', '%H:%M'],
				'x-axis label' => 'Time',
				'y-axis label' => 'Power (W)',
				'extra_opts' => 'set nokey' },
				[{'type' => 'columns', 'style' => 'lines'}, \@seconds, \@all_consumption ] );


	my @used_apps = $self->GetUsedAppliances();

	my $objects;
	my @ytics;

	my $j = 1;
	my @lines;
	for (my $i = 0; $i < scalar(@used_apps); $i++)
	{
		my @slots = $self->GetSlots($used_apps[$i]); 
		push @ytics, ["$used_apps[$i]",  ($i + 1)];
		push @lines, [{'type' => 'function', 'style' => 'lines ls 1'}, $i+1];
		foreach my $slot (@slots)
		{
#print "$used_apps[$i]: @{$slot}\n";
			$objects .= "set object ". $j . " rectangle from " . (@{$slot}[0] * 60) . "," . ($i + 0.7) ." to " . 
						(@{$slot}[1] * 60) . "," . ($i + 1.3) . " fillcolor rgb \"black\" fillstyle solid 1\n";
			$j++;
		}
#$ytics .= ", " if $i < scalar(@used_apps)-1;
	}
#print $objects;

	gnuplot({'title' => $usage_title,
				'output type' => 'eps',
				'output file' => $usage_filename,
				'xdata' => 'time',
				'timefmt' => '%s',
				'format' => ['x', '%H:%M'],
				'x-axis label' => 'Time',
				'y-axis label' => '',
				'yrange' => "[0:".(scalar(@used_apps) + 1)."]",
				'xrange' => "[0:".(1439 * 60)."]",
				'ytics' => \@ytics,
				'extra_opts' => "set style line 1 lt 2 lw 1\nset nokey\n" . $objects },
				@lines );

}

1;
