#    Domestic Lighting Model - Simulation Example Code
#    Original VB script: 
#    Copyright (C) 2009 Ian Richardson, Murray Thomson
#    CREST (Centre for Renewable Energy Systems Technology),
#    Department of Electronic and Electrical Engineering
#    Loughborough University, Leicestershire LE11 3TU, UK
#    Tel. +44 1509 635326. Email address: I.W.Richardson2@lboro.ac.uk

#    Perl implementation:
#    Copyright (C) 2010 Gergely Acs <acs@crysys.hu>

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


package LightingModel;
use strict;
use vars qw($VERSION);
$VERSION = '0.01';

sub new 
{
	my $package = shift;
	return bless({}, $package);
}

sub RunLightingSimulation
{
    my $self = shift;

    my ($light_config, $light_sim_data, $occ_sim_data, $irradiance, $bulbs) = @_; 
 
    my $iTime;

    # Define variables
    my $iLightDuration;
    
    # Initialise the random number generator
    # Randomize

    # Determine the irradiance threshold of this house
    my $iIrradianceThreshold = LightingModel::GetMonteCarloNormalDistGuess($light_config->GlobalIrrMean, 
		    $light_config->GlobalIrrStdDev);

    $light_sim_data->GlobalIrr($iIrradianceThreshold);

    # Clear the target area
    #$light_sim_data->clear_data();
    
    # Choose a random house from the list of 100 provided in the bulbs sheet
    my $iRandomHouse = int(rand() * 100) + 1;
    $light_sim_data->HouseNum($iRandomHouse);
    
    # Declare an array to store the lighting unit configuration
    # vBulbArray            (1,1)   Example dwelling number
    #                       (1,2)   Number of lighting units
    #                       (1,3..)   Lighting unit ratings
    # Get the bulb data
    my @vBulbArray = $bulbs->GetData($iRandomHouse); 
    #Range("bulbs!A" + CStr(iRandomHouse + 10) + ":BI" + CStr(iRandomHouse + 10))
    
    # Get the number of bulbs
    my $iNumBulbs = $vBulbArray[0];
    $light_sim_data->BulbsNum($iNumBulbs);
    # print "\nNumber of bulbs: $iNumBulbs\n\n";
    
    # Declare an array to store the simulation data
    # vSimulation array     (1,n)       Bulb number header
    #                       (2,n)       Rating
    #                       (3,n)       Relative use
    #                       (4-1443,n)  Demand
    my @vSimulationArray; #1 To 1443, 1 To 60) As Variant
    
    # Load the occupancy array
    #my $vOccupancyArray = $occ_sim_data->OccupancyArray; # Range("occ_sim_data!C11:C154")
    
    # Load the irradiance array
    #my $vIrradianceArray = $irradiance->IrradianceArray; # Range("irradiance!C12:C1451")
    
    # Get the calibration scalar
    my $fCalibrationScalar = $light_config->CalibrationScalar; #Range("light_config!F24").Value
    
    # For each bulb
    for (my $i = 0; $i < $iNumBulbs; $i++)
    {
        # Get the bulb rating
        my $iRating = $vBulbArray[$i + 1];
    
        # Store the bulb number
        $vSimulationArray[0][$i] = $i+1;
        
        # Store the rating
        $vSimulationArray[1][$i] = $iRating;
                
        # Assign a random bulb use weighting to this bulb
        # Note that the calibration scalar is multiplied here to save processing time later
        my $fCalibratedRelativeUseWeighting = -$fCalibrationScalar * log(rand());
        $vSimulationArray[2][$i] = $fCalibratedRelativeUseWeighting;
        
        # Calculate the bulb usage at each minute of the day
        $iTime = 0;
        while ($iTime < 1440)
	{
            # Is this bulb switched on to start with?
            # This concept is not implemented in this example.
            # The simplified assumption is that all bulbs are off to start with.
            
            # Get the irradiance for this minute
            my $iIrradiance = $irradiance->GetIrradiance($iTime);
            
            # Get the number of current active occupants for this minute
            # Convert from 10 minute to 1 minute resolution
            my $iActiveOccupants = $occ_sim_data->GetActiveOccupants(int($iTime / 10));
                          
            # Determine if the bulb switch-on condition is passed
            # ie. Insuffient irradiance and at least one active occupant
            # There is a 5% chance of switch on event if the irradiance is above the threshold
            my $bLowIrradiance = $iIrradiance < $iIrradianceThreshold || rand() < 0.05;

            # Get the effective occupancy for this number of active occupants to allow for sharing
            my $fEffectiveOccupancy = $light_config->GetEffectiveOccupancy($iActiveOccupants); 
	    # Range("light_config!E" + CStr(37 + iActiveOccupants)).Value

            # Check the probability of a switch on at this time
	  #  print "fCalibratedRelativeUseWeighting: $fCalibratedRelativeUseWeighting\n";
	  #  print "bLowIrradiance: $bLowIrradiance\n";
	  #  print "iIrradiance: $iIrradiance\n";
	  #  print "iIrradianceThreshold: $iIrradianceThreshold\n";
            if ($bLowIrradiance && rand() < $fEffectiveOccupancy * $fCalibratedRelativeUseWeighting) 
	    {
                # This is a switch on event
		my $r1 = rand();
            
                # Determine how long this bulb is on for
                for (my $j = 1; $j < 10; $j++)
		{ 
                    # Get the cumulative probability of this duration
                    my $cml = $light_config->GetLowerCumProb($j); #("light_config!E" + CStr(54 + j)).Value
                
                    # Check to see if this is the type of light
                    if ($r1 < $cml)
		    {
                        # Get the durations
                        my $iLowerDuration = $light_config->GetLowerMin($j); #"light_config!C" + CStr(54 + j)).Value
                        my $iUpperDuration = $light_config->GetUpperMin($j); #"light_config!D" + CStr(54 + j)).Value
                        
                        # Guess a duration in this range
                        $iLightDuration = (rand() * ($iUpperDuration - $iLowerDuration)) + $iLowerDuration;
			#print "Time: $iTime Bulb: $i Light duration: $iLightDuration\n";
                                                  
                        # Exit the loop
                        last;
		    }
		 }
                 
                 for (my $j = 1; $j <= $iLightDuration; $j++)
		 {
                    # Range check
                    last if ($iTime >= 1440);
                    
                    # Get the number of current active occupants for this minute
                    $iActiveOccupants = $occ_sim_data->GetActiveOccupants(int($iTime / 10));
                    
                    # If there are no active occupants, turn off the light
                    last if $iActiveOccupants == 0;
                    
                    # Store the demand
                    $vSimulationArray[3 + $iTime][$i] = $iRating;
                        
                    # Increment the time
                    $iTime++;
                 } 
	    }
            else
	    {
                # The bulb remains off
                $vSimulationArray[3 + $iTime][$i] = 0;
                
                # Increment the time
                $iTime++;
	    }
	}        
    
    }   # next i 
    # Write the simulation data to the sheet
    $light_sim_data->SimulationArray(\@vSimulationArray);
    #Range("light_sim_data!E9:BL1451") = 
}


sub GetMonteCarloNormalDistGuess #(dMean As Double, dSD As Double) As Integer
{
    # Guess a value from a normal distribution
    # shift;
    my ($dMean, $dSD) = @_;
    my $iGuess;
    my $bOK;
    my $px;
    
    $bOK = $dMean == 0;
    
    $iGuess = 0;

    while (!$bOK)
    {
        # Guess a value
        $iGuess = (rand() * ($dSD * 8)) - ($dSD * 4) + $dMean;

        # See if this is likely
        $px = (1 / ($dSD * ((2 * 3.14159) ** 0.5))) * exp(-(($iGuess - $dMean) ** 2) / (2 * $dSD ** 2));
        
        # End the loop if this value is okay
        $bOK = 1 if ($px >= rand());
    }
    
    # Return the value
    return $iGuess;
}

1;

