#!/usr/bin/perl -w
#    Original VB script: 
#    Copyright (C) 2009 Ian Richardson, Murray Thomson
#    CREST (Centre for Renewable Energy Systems Technology),
#    Department of Electronic and Electrical Engineering
#    Loughborough University, Leicestershire LE11 3TU, UK
#    Tel. +44 1509 635326. Email address: I.W.Richardson2@lboro.ac.uk

#    Perl implementation:
#    Copyright (C) 2010 Gergely Acs <acs@crysys.hu>

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

use Getopt::Std;

use lib './lib';

use strict;

use ApplianceModel;
use LightingModel;
use OccupancyModel;
use Appliance;
use ApplianceSimData;
use OccSimData;
use Bulbs;
use LightSimData;
use LightConfig;
use OccStartStates;
use ActivityStats;
use OccStates;
use Irradiance;

my @month_names = ('January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December');

my $stat_dir = "./stat/";
my $out_dir = "./out/";

mkdir $out_dir;

sub LoadAppliancesFromFile
{
	die("Unable to open $_[0]!\n") unless open (IN, $_[0]);

	my $id = 0;
	my @appliances;

	while (<IN>)
	{
		next if $_ =~ /#.*/;
		my @item = split /[,\n]/;

		#$global_calibration = $item[1], next if $item[0] eq "Calibration";
		next if $item[0] eq "Calibration" || $item[0] eq "Mean active occupancy";
		#$global_mean_active_occ = $item[1], next if $item[0] eq "Mean active occupancy";

		my $appliance = new Appliance();

		$appliance->ID($id++);

		$appliance->Name($item[0]);  
		$appliance->Type($item[16]);  
		$appliance->MeanCycleLength($item[4]); 
		$appliance->CyclesPerYear($item[3]); 
		$appliance->StandbyPower($item[6]); 
		$appliance->RatedPower($item[5]); 
		$appliance->Calibration($item[19]);
		$appliance->Ownership($item[1]); 
		$appliance->TargetAveragekWhYear($item[22]); 
		$appliance->UserProfile($item[17]); 
		$appliance->RestartDelay($item[7]);

	#	print "ID: ", $appliance->ID,"\n";
	#	print "Name: ", $appliance->Name,"\n";
	#	print "Type: ", $appliance->Type,"\n";
	#	print "Mean cycle len: ", $appliance->MeanCycleLength,"\n";
	#	print "Standby power: ", $appliance->StandbyPower,"\n";
	#	print "Rated power: ", $appliance->RatedPower,"\n";
	#	print "Calibration: ", $appliance->Calibration,"\n";
	#	print "Ownership: ", $appliance->Ownership,"\n";
	#	print "Target Avg.Wh.Year: ", $appliance->TargetAveragekWhYear,"\n";
	#	print "User profile: ", $appliance->UserProfile,"\n";
	#	print "Restart delay: ", $appliance->RestartDelay,"\n\n";

		push @appliances, $appliance;
	}
	
	close IN;
	return @appliances;
}

my %args;

getopts("n:r:m:d:",\%args);

die "Usage: simulator.pl -n <number_of_users> [-r <residents_num>] -m <month> -d <we|wd>\n"  
if !defined $args{n} || !defined $args{m} || !defined $args{d};

my ($users_num, $sMonth) = ($args{n}, $args{m});
my $bWeekend = $args{d} eq "we" ? 1 : 0;

my $residentsNum = !defined $args{r} ? int(rand(5) + 1) : $args{r} ;

# Perform a range check on the input month
die "Please give the month of the year from 1 to 12." if ($sMonth < 1 || $sMonth > 12);

die "Please enter the number of residents from 1 to 5." unless ($residentsNum <= 5 && $residentsNum >= 1);

my @appliances = LoadAppliancesFromFile($stat_dir . "appliances.csv");
my $activity_stats = new ActivityStats();
my $bulbs = new Bulbs();
my $irradiance = new Irradiance();
my $light_config = new LightConfig();
my $occ_start_states = new OccStartStates();

# Initialization
$activity_stats->LoadFromFile($stat_dir . "activity_stats.csv");
$bulbs->LoadFromFile($stat_dir . "bulbs.csv");
$irradiance->LoadFromFile($stat_dir . "irradiance.csv");
$light_config->LoadFromFile($stat_dir . "light_config.csv");
$occ_start_states->LoadFromFile($stat_dir . "occ_start_states.csv");

my $Appliance_Model = new ApplianceModel(); 
my $Lighting_Model = new LightingModel(); 
my $Occupancy_Model = new OccupancyModel(); 

foreach my $num  (1..$users_num)
{
# Models
	$residentsNum = int(rand(5) + 1) if !defined $args{r};
	print "Generating dwelling $num [residents: $residentsNum, month: $sMonth, $args{d}]...";

# Data and statistics
	my $light_sim_data = new LightSimData();
	my $occ_sim_data = new OccSimData();
	my $appliance_sim_data = new ApplianceSimData();
	my $occ_states = new OccStates();

	$occ_states->LoadFromFile($stat_dir . "tpm" . $residentsNum . "_" . $args{d} . ".csv");

	$irradiance->Month($sMonth);

# Installing appliances
	$Appliance_Model->ConfigureAppliancesInDwelling(\@appliances);

	# print "\nDwelling appliances:\n\n";

	#print $_->Name(), " : ", ($_->HasAppliance ? "yes" : "no"), "\n" foreach (@appliances);

	# print "\nAppliances have been allocated.\n"; 

# Run the occupancy simulation
	$Occupancy_Model->RunOccupancySimulation($residentsNum, $bWeekend, $occ_start_states, $occ_states, $occ_sim_data);

# Run the lighting simulation
	$Lighting_Model->RunLightingSimulation($light_config, $light_sim_data, $occ_sim_data, $irradiance, $bulbs);

# Run the appliance model simulation
	$Appliance_Model->RunApplianceSimulation(\@appliances, $sMonth, $bWeekend, $Lighting_Model, $occ_sim_data, 
			$appliance_sim_data, $activity_stats);
	my @lighting_consumption = $light_sim_data->GetPowerOnTime();

	$appliance_sim_data->AddLighting(@lighting_consumption);

# Writing simulation data
	$occ_sim_data->WriteToFile($out_dir . "$num-occ_sim_data.csv");
	$appliance_sim_data->WriteToFile($out_dir . "$num-appliance_sim_data.csv", 
			$out_dir . "$num-active_appliances.csv", $out_dir . "$num-all_power.csv", 
			\@lighting_consumption);
	$light_sim_data->WriteToFile($out_dir . "$num-light_sim_data.csv");

# Generating plots
	my $lighting_title = 'Lighting power (' . ($bWeekend ? "weekend, " : "weekday, ") . 
			$month_names[$sMonth-1] .  ", $residentsNum residents) ";
	$light_sim_data->MakePlot($out_dir . "$num-light_sim_data.eps", $lighting_title);

	my $occ_sim_title = 'Number of active occupants (' . 
		($bWeekend ? "weekend, " : "weekday, ") . $month_names[$sMonth-1] .  ", $residentsNum residents) ";
	$occ_sim_data->MakePlot($out_dir . "$num-occ_sim_data.eps", $occ_sim_title);

	my $consumption_title = 'Power (' . ($bWeekend ? "weekend, " : "weekday, ") . $month_names[$sMonth-1] . 
					", $residentsNum residents) ";
	my $usage_title = 'Usage of appliances (' . ($bWeekend ? "weekend, " : "weekday, ") . 
		$month_names[$sMonth-1] .  ", $residentsNum residents) ";
	$appliance_sim_data->MakePlots($out_dir. "$num-all_power.eps", $consumption_title, $out_dir . "$num-usage.eps", $usage_title);

	print "done.\n";

}











